@@ -54,6 +54,9 @@ iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 		return -1;
 	}
 
+	pdu->callback     = cb;
+	pdu->private_data = private_data;
+
 	/* flags */
 	iscsi_pdu_set_pduflags(pdu, 0x80);
 
@@ -64,26 +67,22 @@ iscsi_nop_out_async(struct iscsi_context *iscsi, iscsi_command_cb cb,
 	iscsi_pdu_set_lun(pdu, 0);
 
 	/* cmdsn */
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
 	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
 
-	pdu->callback     = cb;
-	pdu->private_data = private_data;
-
 	if (data != NULL && len > 0) {
 		if (iscsi_pdu_add_data(iscsi, pdu, data, len) != 0) {
+                        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 			iscsi_set_error(iscsi, "Failed to add outdata to nop-out");
 			iscsi->drv->free_pdu(iscsi, pdu);
 			return -1;
 		}
 	}
-
-	if (iscsi_queue_pdu(iscsi, pdu) != 0) {
-		iscsi_set_error(iscsi, "failed to queue iscsi nop-out pdu");
-		iscsi->drv->free_pdu(iscsi, pdu);
-		return -1;
-	}
-
 	iscsi->cmdsn++;
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
+
+	iscsi_queue_pdu(iscsi, pdu);
+
 	iscsi->nops_in_flight++;
 	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
 	          "NOP Out Send (nops_in_flight: %d, pdu->cmdsn %08x, pdu->itt %08x, pdu->ttt %08x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x)",
@@ -122,11 +121,7 @@ iscsi_send_target_nop_out(struct iscsi_context *iscsi, uint32_t ttt, uint32_t lu
 	/* cmdsn is not increased if Immediate delivery*/
 	iscsi_pdu_set_cmdsn(pdu, iscsi->cmdsn);
 
-	if (iscsi_queue_pdu(iscsi, pdu) != 0) {
-		iscsi_set_error(iscsi, "failed to queue iscsi nop-out pdu");
-		iscsi->drv->free_pdu(iscsi, pdu);
-		return -1;
-	}
+	iscsi_queue_pdu(iscsi, pdu);
 
 	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
 	          "NOP Out Send (nops_in_flight: %d, pdu->cmdsn %08x, pdu->itt %08x, pdu->ttt %08x, pdu->lun %8x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x)",
@@ -141,6 +136,7 @@ iscsi_process_nop_out_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 {
 	struct iscsi_data data;
 
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
 	ISCSI_LOG(iscsi, (iscsi->nops_in_flight > 1) ? 1 : 6,
 	          "NOP-In received (pdu->itt %08x, pdu->ttt %08x, iscsi->maxcmdsn %08x, iscsi->expcmdsn %08x, iscsi->statsn %08x)",
 	          pdu->itt, 0xffffffff, iscsi->maxcmdsn, iscsi->expcmdsn, iscsi->statsn); 
@@ -155,6 +151,7 @@ iscsi_process_nop_out_reply(struct iscsi_context *iscsi, struct iscsi_pdu *pdu,
 		iscsi->nops_in_flight = 0;
 	}
 	iscsi->min_cmdsn_waiting = iscsi->waitpdu->cmdsn;
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 
 	if (pdu->callback == NULL) {
 		return 0;
