@@ -62,6 +62,11 @@
 #include <sys/uio.h>
 #endif
 
+#ifdef HAVE_PTHREAD
+#include <signal.h>
+#endif
+
+#include <stdbool.h>
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -87,7 +92,7 @@ union socket_address {
 void
 iscsi_add_to_outqueue(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 {
-	struct iscsi_pdu *current = iscsi->outqueue;
+	struct iscsi_pdu *current;
 	struct iscsi_pdu *last = NULL;
 
 	if (iscsi->scsi_timeout > 0) {
@@ -96,12 +101,15 @@ iscsi_add_to_outqueue(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 		pdu->scsi_timeout = 0;
 	}
 
-	if (iscsi->outqueue == NULL) {
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
+
+        current = iscsi->outqueue;
+        if (iscsi->outqueue == NULL) {
 		iscsi->outqueue = pdu;
 		pdu->next = NULL;
-		return;
+                goto finished;
 	}
-	
+        
 	/* queue pdus in ascending order of CmdSN. 
 	 * ensure that pakets with the same CmdSN are kept in FIFO order.
 	 * immediate PDUs are queued in front of queue with the CmdSN
@@ -125,22 +133,42 @@ iscsi_add_to_outqueue(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
 			(pdu->outdata.data[0] & ISCSI_PDU_IMMEDIATE && !(current->outdata.data[0] & ISCSI_PDU_IMMEDIATE))) {
 			/* insert PDU before the current */
 			if (last != NULL) {
-				last->next=pdu;
+				last->next = pdu;
 			} else {
-				iscsi->outqueue=pdu;
+				iscsi->outqueue = pdu;
 			}
 			pdu->next = current;
-			return;
+                        goto finished;
 		}
-		last=current;
-		current=current->next;
+		last = current;
+		current = current->next;
 	} while (current != NULL);
 	
 	last->next = pdu;
 	pdu->next = NULL;
+
+ finished:
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
+        
+#if defined(HAVE_MULTITHREADING) && defined(HAVE_PTHREAD)
+        if(iscsi->multithreading_enabled) {
+                if (current == NULL && pdu == iscsi->outqueue) {
+                        pthread_kill(iscsi->service_thread, SIGUSR1);
+                }
+        } else {
+#endif
+                if (iscsi->outqueue == pdu) {
+                        iscsi->drv->service(iscsi, POLLOUT);
+                }
+#if defined(HAVE_MULTITHREADING) && defined(HAVE_PTHREAD)
+        }
+#endif
+
+        return;
 }
 
 void iscsi_decrement_iface_rr() {
+        /* TODO QQQ use an atomic here */
 	iface_rr--;
 }
 
@@ -305,6 +333,7 @@ iscsi_connect_async(struct iscsi_context *iscsi, const char *portal,
 	struct addrinfo *ai = NULL;
 	union socket_address sa;
 	int socksize;
+	bool portal_is_ip;
 
 	ISCSI_LOG(iscsi, 2, "connecting to portal %s",portal);
 
@@ -351,6 +380,10 @@ iscsi_connect_async(struct iscsi_context *iscsi, const char *portal,
 		*str = 0;
 	}
 
+        /* check if we got an ip address or hostname for portal */
+	portal_is_ip = inet_pton(AF_INET, host, &sa) == 1 ||
+		       inet_pton(AF_INET6, host, &sa) == 1;
+
 	/* is it a hostname ? */
 	if (getaddrinfo(host, NULL, NULL, &ai) != 0) {
 		iscsi_free(iscsi, addr);
@@ -370,6 +403,11 @@ iscsi_connect_async(struct iscsi_context *iscsi, const char *portal,
 #ifdef HAVE_SOCK_SIN_LEN
 		sa.sin.sin_len = socksize;
 #endif
+		if (!portal_is_ip) {
+			char ip[INET_ADDRSTRLEN];
+			inet_ntop(AF_INET, &sa.sin.sin_addr, ip, sizeof(ip));
+			ISCSI_LOG(iscsi, 2, "portal resolved to ipv4 address %s", ip);
+		}
 		break;
 #ifdef HAVE_SOCKADDR_IN6
 	case AF_INET6:
@@ -380,6 +418,11 @@ iscsi_connect_async(struct iscsi_context *iscsi, const char *portal,
 #ifdef HAVE_SOCK_SIN_LEN
 		sa.sin6.sin6_len = socksize;
 #endif
+		if (!portal_is_ip) {
+			char ip[INET6_ADDRSTRLEN];
+			inet_ntop(AF_INET6, &sa.sin6.sin6_addr, ip, sizeof(ip));
+			ISCSI_LOG(iscsi, 2, "portal resolved to ipv6 address %s", ip);
+		}
 		break;
 #endif
 	default:
@@ -415,7 +458,11 @@ iscsi_tcp_disconnect(struct iscsi_context *iscsi)
 		return -1;
 	}
 
-	close(iscsi->fd);
+	if (iscsi->old_iscsi && iscsi->old_iscsi->fd == iscsi->fd) {
+		/* Reserve this fd because old_iscsi->fd will be reused */
+	} else {
+		close(iscsi->fd);
+	}
 
 	if (!(iscsi->pending_reconnect && iscsi->old_iscsi) &&
 	    iscsi->connected_portal[0]) {
@@ -464,14 +511,16 @@ iscsi_tcp_which_events(struct iscsi_context *iscsi)
 		return 0;
 	}
 
-	if (iscsi->outqueue_current != NULL ||
-	    (iscsi->outqueue != NULL && !iscsi->is_corked &&
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
+	if (iscsi->outqueue_current ||
+	    (iscsi->outqueue && !iscsi->is_corked &&
 	     (iscsi_serial32_compare(iscsi->outqueue->cmdsn, iscsi->maxcmdsn) <= 0 ||
 	      iscsi->outqueue->outdata.data[0] & ISCSI_PDU_IMMEDIATE)
 	    )
 	   ) {
 		events |= POLLOUT;
 	}
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 	return events;
 }
 
@@ -487,6 +536,7 @@ iscsi_queue_length(struct iscsi_context *iscsi)
 	int i = 0;
 	struct iscsi_pdu *pdu;
 
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
 	for (pdu = iscsi->outqueue; pdu; pdu = pdu->next) {
 		i++;
 	}
@@ -496,6 +546,7 @@ iscsi_queue_length(struct iscsi_context *iscsi)
 	if (iscsi->is_connected == 0) {
 		i++;
 	}
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 
 	return i;
 }
@@ -506,15 +557,17 @@ iscsi_out_queue_length(struct iscsi_context *iscsi)
 	int i = 0;
 	struct iscsi_pdu *pdu;
 
+        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
 	for (pdu = iscsi->outqueue; pdu; pdu = pdu->next) {
 		i++;
 	}
+        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 
 	return i;
 }
 
 ssize_t
-iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *iovector, uint32_t pos, ssize_t count, int do_write)
+iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *iovector, uint32_t pos, ssize_t count, uint32_t *data_digest_ptr, int do_write)
 {
 	struct scsi_iovec *iov, *iov2;
 	int niov;
@@ -598,6 +651,19 @@ iscsi_iovector_readv_writev(struct iscsi_context *iscsi, struct scsi_iovector *i
 		n = readv(iscsi->fd, (struct iovec*) iov, niov);
 	}
 
+	/* Update the data digest */
+	if (data_digest_ptr && n > 0) {
+		int i;
+		size_t bytes_to_crc = n;
+		struct iovec *iov_ptr = (struct iovec*)iov;
+
+		for ( i=0; iov_ptr && i<niov && bytes_to_crc; iov_ptr++, i++) {
+			size_t chunk = MIN(bytes_to_crc, iov_ptr->iov_len);
+			*data_digest_ptr = crc32c_chain(*data_digest_ptr, iov_ptr->iov_base, chunk);
+			bytes_to_crc -= chunk;
+		}
+	}
+
 	/* restore original values */
 	iov->iov_base = (void*) ((uintptr_t)iov->iov_base - pos);
 	iov->iov_len += pos;
@@ -618,20 +684,25 @@ static int
 iscsi_read_from_socket(struct iscsi_context *iscsi)
 {
 	struct iscsi_in_pdu *in;
-	ssize_t hdr_size, data_size, count, padding_size;
-
+	ssize_t hdr_size, data_size, count, padding_size, data_segment_len;
+	bool do_data_digest = (iscsi->data_digest != ISCSI_DATA_DIGEST_NONE);
+        int ret = -1;
+        
 	do {
 		hdr_size = ISCSI_HEADER_SIZE(iscsi->header_digest);
 		if (iscsi->incoming == NULL) {
-			iscsi->incoming = iscsi_szmalloc(iscsi, sizeof(struct iscsi_in_pdu));
+			iscsi->incoming = iscsi_zmalloc(iscsi, sizeof(struct iscsi_in_pdu));
 			if (iscsi->incoming == NULL) {
 				iscsi_set_error(iscsi, "Out-of-memory: failed to malloc iscsi_in_pdu");
-				return -1;
+                                goto finished;
 			}
-			iscsi->incoming->hdr = iscsi_smalloc(iscsi, hdr_size);
+                }
+		if (iscsi->incoming->hdr == NULL) {
+			crc32c_init(&(iscsi->incoming->calculated_data_digest));
+			iscsi->incoming->hdr = iscsi_malloc(iscsi, hdr_size);
 			if (iscsi->incoming->hdr == NULL) {
 				iscsi_set_error(iscsi, "Out-of-memory");
-				return -1;
+                                goto finished;
 			}
 		}
 		in = iscsi->incoming;
@@ -646,7 +717,7 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
                                      count, 0);
 			if (count == 0) {
 				/* remote side has closed the socket. */
-				return -1;
+                                goto finished;
 			}
 			if (count < 0) {
 				if (errno == EINTR || errno == EAGAIN) {
@@ -654,7 +725,7 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
 				}
 				iscsi_set_error(iscsi, "read from socket failed, "
 					"errno:%d", errno);
-				return -1;
+                                goto finished;
 			}
 			in->hdr_pos  += count;
 		}
@@ -665,11 +736,12 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
 		}
 
 		padding_size = iscsi_get_pdu_padding_size(&in->hdr[0]);
-		data_size = iscsi_get_pdu_data_size(&in->hdr[0]) + padding_size;
+		data_segment_len = iscsi_get_pdu_data_size(&in->hdr[0]);
+		data_size = data_segment_len + padding_size;
 
-		if (data_size < 0 || data_size > (ssize_t)iscsi->initiator_max_recv_data_segment_length) {
-			iscsi_set_error(iscsi, "Invalid data size received from target (%d)", (int)data_size);
-			return -1;
+		if (data_segment_len < 0 || data_segment_len > (ssize_t)iscsi->initiator_max_recv_data_segment_length) {
+			iscsi_set_error(iscsi, "Invalid datasegmentlen received from target (%d)", (int)data_segment_len);
+                        goto finished;
 		}
 		if (data_size != 0) {
 			unsigned char padding_buf[3];
@@ -682,29 +754,31 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
 			iovector_in = iscsi_get_scsi_task_iovector_in(iscsi, in);
 			if (iovector_in != NULL && count > padding_size) {
 				uint32_t offset = scsi_get_uint32(&in->hdr[40]);
-				count = iscsi_iovector_readv_writev(iscsi, iovector_in, in->data_pos + offset, count - padding_size, 0);
+				count = iscsi_iovector_readv_writev(iscsi, iovector_in, in->data_pos + offset, count - padding_size, do_data_digest ? &(in->calculated_data_digest) : NULL, 0);
 			} else {
 				if (iovector_in == NULL) {
 					if (in->data == NULL) {
 						in->data = iscsi_malloc(iscsi, data_size);
 						if (in->data == NULL) {
 							iscsi_set_error(iscsi, "Out-of-memory: failed to malloc iscsi_in_pdu->data(%d)", (int)data_size);
-							return -1;
+                                                        goto finished;
 						}
 					}
 					buf = &in->data[in->data_pos];
 				}
 				count = recv(iscsi->fd, (void *)buf, count, 0);
+				if (do_data_digest && count > 0)
+					in->calculated_data_digest = crc32c_chain(in->calculated_data_digest, buf, count);
 			}
 			if (count == 0) {
 				/* remote side has closed the socket. */
-				return -1;
+                                goto finished;
 			}
 			if (count < 0) {
 				if (errno == EINTR || errno == EAGAIN) {
 					break;
 				}
-				return -1;
+                                goto finished;
 			}
 			in->data_pos += count;
 		}
@@ -713,15 +787,39 @@ iscsi_read_from_socket(struct iscsi_context *iscsi)
 			break;
 		}
 
-		iscsi->incoming = NULL;
+		/* Handle Data Digest receive */
+		if (data_size != 0 && do_data_digest &&
+			in->received_data_digest_bytes < ISCSI_DIGEST_SIZE) {
+
+			count = recv(iscsi->fd, (void *)(in->data_digest_buf + in->received_data_digest_bytes), ISCSI_DIGEST_SIZE - in->received_data_digest_bytes, 0);
+			if (count == 0) {
+				/* remote side has closed the socket. */
+                                goto finished;
+			}
+			if (count < 0) {
+				if (errno == EINTR || errno == EAGAIN) {
+					break;
+				}
+                                goto finished;
+			}
+			in->received_data_digest_bytes += count;
+
+			if (in->received_data_digest_bytes < ISCSI_DIGEST_SIZE) {
+				break;
+			}
+		}
+
+                iscsi->incoming = NULL;
 		if (iscsi_process_pdu(iscsi, in) != 0) {
 			iscsi_free_iscsi_in_pdu(iscsi, in);
-			return -1;
+                        goto finished;
 		}
 		iscsi_free_iscsi_in_pdu(iscsi, in);
-	} while (iscsi->tcp_nonblocking && iscsi->waitpdu && iscsi->is_loggedin);
+        } while (iscsi->tcp_nonblocking && iscsi->waitpdu && iscsi->is_loggedin); //QQQ break the loop
 
-	return 0;
+        ret = 0;
+ finished:
+	return ret;
 }
 
 static int iscsi_pdu_update_headerdigest(struct iscsi_context *iscsi, struct iscsi_pdu *pdu)
@@ -746,12 +844,12 @@ static int iscsi_pdu_update_headerdigest(struct iscsi_context *iscsi, struct isc
 static int
 iscsi_write_to_socket(struct iscsi_context *iscsi)
 {
-	ssize_t count;
+	ssize_t count, data_segment_len;
 	size_t total;
 	struct iscsi_pdu *pdu;
 	static char padding_buf[3];
 	int socket_flags = 0;
-
+	bool do_data_digest = (iscsi->data_digest != ISCSI_DATA_DIGEST_NONE), execute_data_digest = false;
 #ifdef MSG_NOSIGNAL
 	socket_flags |= MSG_NOSIGNAL;
 #elif SO_NOSIGPIPE
@@ -763,7 +861,7 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 		return -1;
 	}
 
-	while (iscsi->outqueue != NULL || iscsi->outqueue_current != NULL) {
+	while (iscsi->outqueue || iscsi->outqueue_current) {
 		if (iscsi->outqueue_current == NULL) {
 			if (iscsi->is_corked) {
 				/* connection is corked we are not allowed to send
@@ -788,14 +886,20 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 				                iscsi->outqueue->cmdsn, iscsi->expcmdsn, iscsi->outqueue->outdata.data[0] & 0x3f);
 				return -1;
 			}
+                        
+                        iscsi_mt_spin_lock(&iscsi->iscsi_lock);
 			iscsi->outqueue_current = iscsi->outqueue;
 			
 			/* set exp statsn */
-			iscsi_pdu_set_expstatsn(iscsi->outqueue_current, iscsi->statsn + 1);
+			if((iscsi->outqueue->outdata.data[0] & 0x3f) != ISCSI_PDU_DATA_OUT)
+				iscsi_pdu_set_expstatsn(iscsi->outqueue_current, iscsi->statsn + 1);
+			else
+				iscsi_pdu_set_expstatsn(iscsi->outqueue_current, iscsi->statsn);
 
 			/* calculate header checksum */
 			if (iscsi->header_digest != ISCSI_HEADER_DIGEST_NONE &&
 				iscsi_pdu_update_headerdigest(iscsi, iscsi->outqueue_current) != 0) {
+                                iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 				return -1;
 			}
 
@@ -807,6 +911,7 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 				   cmd PDU the R2T might get lost otherwise. */
 				ISCSI_LIST_ADD_END(&iscsi->waitpdu, iscsi->outqueue_current);
 			}
+                        iscsi_mt_spin_unlock(&iscsi->iscsi_lock);
 		}
 
 		pdu = iscsi->outqueue_current;
@@ -834,6 +939,24 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 			return 0;
 		}
 
+		
+		if (do_data_digest) {
+			data_segment_len = iscsi_get_pdu_data_size(pdu->outdata.data);
+
+			if (data_segment_len && !pdu->payload_len)
+				execute_data_digest = true;
+			else
+				execute_data_digest = false;
+		}
+
+		if (execute_data_digest && !pdu->outdata_digest_computed) {
+			uint8_t ahslen = pdu->outdata.data[4];
+			uint8_t head_len = iscsi->header_digest != ISCSI_HEADER_DIGEST_NONE ? 52 : 48;
+			uint32_t offset = head_len + ahslen * 4;
+			pdu->calculated_data_digest = crc32c_chain(pdu->calculated_data_digest, pdu->outdata.data + offset, pdu->outdata.size - offset);
+			pdu->outdata_digest_computed = true;
+		}
+
 		/* Write any iovectors that might have been passed to us */
 		while (pdu->payload_written < pdu->payload_len) {
 			struct scsi_iovector* iovector_out;
@@ -848,7 +971,7 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 			count = iscsi_iovector_readv_writev(iscsi,
 				iovector_out,
 				pdu->payload_offset + pdu->payload_written,
-				pdu->payload_len - pdu->payload_written, 1);
+				pdu->payload_len - pdu->payload_written, do_data_digest ? &(pdu->calculated_data_digest) : NULL, 1);
 			if (count == -1) {
 				if (errno == EAGAIN || errno == EWOULDBLOCK) {
 					return 0;
@@ -873,12 +996,56 @@ iscsi_write_to_socket(struct iscsi_context *iscsi)
 						"socket :%d", errno);
 				return -1;
 			}
+
+			if (do_data_digest)
+				pdu->calculated_data_digest = crc32c_chain(pdu->calculated_data_digest, (uint8_t *)padding_buf, count);
+
 			pdu->payload_written += count;
 		}
+
 		/* if we havent written the full padding yet. */
+		if (pdu->payload_written < total) {
+			return 0;
+		}
+
+		/*
+		 * Maybe update the total again, and write the digest, but only if
+		 * 1. DataDigest has been negociated, and
+		 * 2. We have actually written some data
+		 */
+		if (execute_data_digest || (do_data_digest && pdu->payload_written)) {
+			uint32_t data_digest = crc32c_chain_done(pdu->calculated_data_digest);
+			char data_digest_buf[ISCSI_DIGEST_SIZE];
+
+			total += ISCSI_DIGEST_SIZE;
+
+			data_digest_buf[3] = (data_digest >> 24);
+			data_digest_buf[2] = (data_digest >> 16);
+			data_digest_buf[1] = (data_digest >>  8);
+			data_digest_buf[0] = (data_digest);
+
+			/* Write data digest */
+			if (pdu->payload_written < total) {
+				int todo = total - pdu->payload_written;
+				count = send(iscsi->fd, data_digest_buf + (ISCSI_DIGEST_SIZE - todo), todo, socket_flags);
+				if (count == -1) {
+					if (errno == EAGAIN || errno == EWOULDBLOCK) {
+						return 0;
+					}
+					iscsi_set_error(iscsi, "Error when writing to "
+							"socket :%d", errno);
+					return -1;
+				}
+
+				pdu->payload_written += count;
+			}
+		}
+
+		/* if we havent written everything yet. */
 		if (pdu->payload_written != total) {
 			return 0;
 		}
+
 		if (pdu->flags & ISCSI_PDU_CORK_WHEN_SENT) {
 			iscsi->is_corked = 1;
 		}
@@ -984,8 +1151,10 @@ iscsi_tcp_service(struct iscsi_context *iscsi, int revents)
 		}
 
 		if (getsockname(iscsi->fd, (struct sockaddr *) &local, &local_l) == 0) {
-			ISCSI_LOG(iscsi, 2, "connection established (%s:%u -> %s)", inet_ntoa(local.sin_addr),
-						(unsigned)ntohs(local.sin_port),iscsi->connected_portal);
+			char ip[INET_ADDRSTRLEN];
+			inet_ntop(AF_INET, &local.sin_addr, ip, sizeof(ip));
+			ISCSI_LOG(iscsi, 2, "connection established (%s:%u -> %s)", ip,
+			          (unsigned)ntohs(local.sin_port), iscsi->connected_portal);
 		}
 
 		iscsi->is_connected = 1;
@@ -1003,6 +1172,7 @@ iscsi_tcp_service(struct iscsi_context *iscsi, int revents)
 			return iscsi_service_reconnect_if_loggedin(iscsi);
 		}
 	}
+        
 	if (revents & POLLOUT) {
 		if (iscsi_write_to_socket(iscsi) != 0) {
 			ISCSI_LOG(iscsi, 1, "%s", iscsi_get_error(iscsi));
@@ -1026,26 +1196,19 @@ iscsi_service(struct iscsi_context *iscsi, int revents)
 	return iscsi->drv->service(iscsi, revents);
 }
 
-static int iscsi_tcp_queue_pdu(struct iscsi_context *iscsi,
+static void iscsi_tcp_queue_pdu(struct iscsi_context *iscsi,
                                struct iscsi_pdu *pdu)
 {
-	if (pdu == NULL) {
-		iscsi_set_error(iscsi, "trying to queue NULL pdu");
-		return -1;
-	}
-
 	iscsi_add_to_outqueue(iscsi, pdu);
-
-	return 0;
 }
 
 void
 iscsi_free_iscsi_in_pdu(struct iscsi_context *iscsi, struct iscsi_in_pdu *in)
 {
-	iscsi_sfree(iscsi, in->hdr);
+	iscsi_free(iscsi, in->hdr);
 	iscsi_free(iscsi, in->data);
 	in->data=NULL;
-	iscsi_sfree(iscsi, in);
+	iscsi_free(iscsi, in);
 	in=NULL;
 }
 
